searchNodes=[{"doc":"Documentation for Opus. see Opus.Pipeline .","ref":"Opus.html","title":"Opus","type":"module"},{"doc":"Defines a pipeline. A pipeline defines a single entry point function to start running the defined stages. A simple pipeline module can be defined as: defmodule ArithmeticPipeline do use Opus.Pipeline step :to_integer , with : &amp; :erlang . binary_to_integer / 1 step :double , with : &amp; &amp;1 * 2 end # Invoke it with: ArithmeticPipeline . call &quot;42&quot; # =&gt; {:ok, 84} The pipeline can be run calling a call/1 function which is defined by Opus.Pipeline . Pipelines are intended to have a single parameter and always return a tagged tuple {:ok, value} | {:error, error} . A stage returning {:error, error} halts the pipeline. The error value is an Opus.PipelineError struct which contains useful information to detect where the error was caused and why. Stage Definition The following types of stages can be defined: step - see Opus.Pipeline.Stage.Step tee - see Opus.Pipeline.Stage.Tee check - see Opus.Pipeline.Stage.Check link - see Opus.Pipeline.Stage.Link skip - see Opus.Pipeline.Stage.Skip Stage Options :with : The function to call to fulfill this stage. It can be an Atom referring to a public function of the module, an anonymous function or a function reference. :if : Makes a stage conditional, it can be either an Atom referring to a public function of the module, an anonymous function or a function reference. For the stage to be executed, the condition must return true . When the stage is skipped, the input is forwarded to the next step if there's one. :unless : The opposite of the :if option, executes the step only when the callback function returns false . :raise : A list of exceptions to not rescue. Defaults to false which converts all exceptions to {:error, %Opus.PipelineError{}} values halting the pipeline. :error_message : An error message to replace the original error when a stage fails. It can be a String or Atom, which will be used directly in place of the original message, or an anonymous function, which receives the input of the failed stage and must return the error message to be used. :retry_times : How many times to retry a failing stage, before halting the pipeline. :retry_backoff : A backoff function to provide delay values for retries. It can be an Atom referring to a public function in the module, an anonymous function or a function reference. It must return an Enumerable.t yielding at least as many numbers as the retry_times . :instrument? : A boolean which defaults to true . Set to false to skip instrumentation for a stage. Exception Handling All exceptions are converted to {:error, exception} tuples by default. You may let a stage raise an exception by providing the :raise option to a stage as follows: defmodule ArithmeticPipeline do use Opus.Pipeline step :to_integer , &amp; :erlang . binary_to_integer / 1 , raise : [ ArgumentError ] end Stage Filtering You can select the stages of a pipeline to run using call/2 with the :except and :only options. Example: # Runs only the stage with the :validate_params name CreateUserPipeline . call ( params , only : [ :validate_params ] # Runs all the stages except the selected ones CreateUserPipeline . call ( params , except : :send_notification )","ref":"Opus.Pipeline.html","title":"Opus.Pipeline","type":"module"},{"doc":"","ref":"Opus.Pipeline.html#check/2","title":"Opus.Pipeline.check/2","type":"macro"},{"doc":"","ref":"Opus.Pipeline.html#link/2","title":"Opus.Pipeline.link/2","type":"macro"},{"doc":"","ref":"Opus.Pipeline.html#skip/2","title":"Opus.Pipeline.skip/2","type":"macro"},{"doc":"","ref":"Opus.Pipeline.html#step/2","title":"Opus.Pipeline.step/2","type":"macro"},{"doc":"","ref":"Opus.Pipeline.html#tee/2","title":"Opus.Pipeline.tee/2","type":"macro"},{"doc":"","ref":"Opus.Pipeline.html#t:opts/0","title":"Opus.Pipeline.opts/0","type":"type"},{"doc":"","ref":"Opus.Pipeline.html#t:result/0","title":"Opus.Pipeline.result/0","type":"type"},{"doc":"The check stage is intended to assert its input fulfils certain criteria otherwise the pipeline is halted. Its implementation must return true to proceed running any next steps. When the check fails, an Atom error message is provided by default to allow for pattern-matching. defmodule CreateUserPipeline do use Opus.Pipeline check :valid_params? , with : &amp; UserValidator . validate / 1 end When the :valid_params? check fails, the return value of the pipeline will be: {:error, %Opus.PipelineError{error: :failed_check_valid_params?}}","ref":"Opus.Pipeline.Stage.Check.html","title":"Opus.Pipeline.Stage.Check","type":"module"},{"doc":"The link stage calls the specified pipeline module. When defined with a non Opus.Pipeline module, it ignores it. Example defmodule AddOnePipeline do use Opus.Pipeline step :add , with : &amp; ( &amp;1 + 1 ) end defmodule MultiplicationPipeline do use Opus.Pipeline step :double , with : &amp; ( &amp;1 * 2 ) link AddOnePipeline end MultiplicationPipeline . call 5 # {:ok, 11}","ref":"Opus.Pipeline.Stage.Link.html","title":"Opus.Pipeline.Stage.Link","type":"module"},{"doc":"The skip stage is meant to halt the pipeline with no error if the given condition is true. This stage must always be defined with an if option, in order to decide if the pipeline is going to be halted or not. When the given conditional is true , the pipeline will return {:ok, :skipped} and all the following steps will be skipped. defmodule CreateUserPipeline do use Opus.Pipeline skip :prevent_duplicates , if : :user_exists? step :persist_user def user_exists? ( _ ) , do : &quot;implementation omitted&quot; end In this example, if the user_exists? implementation returns true , then the next step persist_user is not going to be called. If false or any other value, then Opus will keep following to the next stages.","ref":"Opus.Pipeline.Stage.Skip.html","title":"Opus.Pipeline.Stage.Skip","type":"module"},{"doc":"The step stage defines an operation which is considered successful unless it returns either an error atom :error or tuple {:error, _} . It is also considered failed and halts the pipeline when it raises an unexpected exception. Example defmodule CryptoMarkerForecastPipeline do use Opus.Pipeline step :waste_time , with : ( fn _ -&gt; Process . sleep ( 10 ) end ) step :calculate_lunar_phase step :fetch_elonmusks_tweets step :forecast # Step definitions can either be defined inline using the `with` option # or as module functions like below # Notice that all step functions expect a single argument. # The return value of a step becomes the input value of the next one. def calculate_lunar_phase ( _ ) do [ &quot;🌑&quot; , &quot;🌒&quot; , &quot;🌓&quot; , &quot;🌔&quot; , &quot;🌖&quot; , &quot;🌗&quot; , &quot;🌘&quot; , &quot;🌚&quot; , &quot;🌜&quot; , &quot;🌝&quot; ] |&gt; Enum . random end def fetch_elonmusks_tweets ( _ ) , do : &quot;Baby Doge, doo, doo, doo&quot; def forecast ( _ ) do if :random . uniform &gt; 0.5 do :buy else :sell end end end The above pipeline module can be invoked with: CryptoMarkerForecastPipeline . call &quot;anything&quot;","ref":"Opus.Pipeline.Stage.Step.html","title":"Opus.Pipeline.Stage.Step","type":"module"},{"doc":"The tee stage is intended for side-effects with no meaningful return value. Its return value is ignored and the next stage is always called.","ref":"Opus.Pipeline.Stage.Tee.html","title":"Opus.Pipeline.Stage.Tee","type":"module"},{"doc":"Module to refine the stages to be run in a pipeline. Options :except : A list of names, or an atom of stages to skip :only : A list of names, or an atom of stages to keep","ref":"Opus.Pipeline.StageFilter.html","title":"Opus.Pipeline.StageFilter","type":"module"},{"doc":"Emits telemetry events. To enable this instrumentation module, update your config/config.exs file with: config :opus , :instrumentation , [ Opus.Telemetry ] Telemetry Events [:opus, :pipeline, :start] - emitted when a pipeline module is called. Measurement: %{time: System.system_time()} Metadata: %{pipeline: String.t()} [:opus, :pipeline, :stage, :stop] - emitted at the end of a stage. Measurement: %{duration: pos_integer()} Metadata: %{pipeline: String.t(), stage: String.t()} [:opus, :pipeline, :stop] - emitted when a pipeline has been completed. Measurement: %{duration: pos_integer(), success: boolean()} Metadata: %{pipeline: String.t()}","ref":"Opus.Telemetry.html","title":"Opus.Telemetry","type":"module"},{"doc":"","ref":"Opus.Telemetry.html#instrument/3","title":"Opus.Telemetry.instrument/3","type":"function"},{"doc":"Error struct capturing useful information to detect where an error was caused and why.","ref":"Opus.PipelineError.html","title":"Opus.PipelineError","type":"exception"},{"doc":"Callback implementation for Exception.message/1 .","ref":"Opus.PipelineError.html#message/1","title":"Opus.PipelineError.message/1","type":"function"},{"doc":"","ref":"Opus.PipelineError.html#t:t/0","title":"Opus.PipelineError.t/0","type":"type"},{"doc":"# Change Log All notable changes to this project will be documented in this file. This project adheres to [Semantic Versioning](http://semver.org/). ## Unreleased ## [0.8.4] - 2022-02-25 ### Changed * Used reactive Kino.Inputs in tutorial live notebook ## [0.8.3] - 2021-11-03 ### Changed * Correct tutorial livebook header ## [0.8.2] - 2021-11-03 ### Added * A tutorial livebook documentation ## [0.8.1] - 2021-09-29 ### Changed * Depend on telemetry@1.0 where possible ## [0.8.0] - 2021-07-17 ### Changed * You can now pass anonymous functions or a function reference to `error_message`. ## [0.7.0] - 2021-04-08 ### Added * `Opus.Telemetry` instrumentation module which emits `:telemetry` events. Read [more][post-opus-telemetry]. ## [0.6.0] - 2019-04-13 ### Added * `skip` macro to bypass a linked pipeline module. See: https://github.com/zorbash/opus#skip ## [0.5.3] - 2019-03-10 ### Added * `:unless` option to conditionally skip a stage ## [0.5.2] - 2019-02-27 ### Changed * Removed `graphvix` from the mix file ### Changed * Opus.Graph has been extracted to a separate package. See: https://github.com/zorbash/opus_graph ## [0.5.1] - 2018-05-08 ## Fixed * Raise CompileError when an unknown module is linked. Previous behaviour was that an unknown module would be silently ignored. ## [0.5.0] - 2018-05-08 ### Added * Module-Global options (see: https://github.com/zorbash/opus#module-global-options) * Opus.PipelineError struct now includes a stacktrace (when it's a wrapped exception) * New instrumentation events, :pipeline_started and :pipeline_completed (holds total time for a pipeline) ## [0.4.0] - 2018-05-03 ### Added * Opus.Graph.generate/1 to create visualisations of pipelines of an app. Read more [here]( https://github.com/zorbash/opus#graph). ## Changes * pipeline?/0 stages/0 and instrument/3 of pipeline modules are now hidden from the docs ## [0.3.1] - 2018-04-10 ### Fixed * The :error_message option is respected in check stages and the default error message is an Atom, constructed :&quot;failed_check_\#{stage_name}&quot;. ## [0.3.0] - 2018-04-02 ### Added * `call/2` to pipeline modules with `:except` and `:only` options ## [0.2.0] - 2018-03-31 ### Added * `retry_times` and `retry_backoff` options * `error_message` option ## [0.1.1] - 2018-03-21 ### Fixed * Bug where recursive pipeline linking would fail (see: 07a44488f8f) ## [0.1.0] - 2018-03-05 ### Added * Initial release [post-opus-telemetry]: https://zorbash.com/post/phoenix-telemetry/","ref":"changelog.html","title":"CHANGELOG","type":"extras"},{"doc":"A framework for pluggable business logic components.","ref":"readme.html","title":"Opus","type":"extras"},{"doc":"The package can be installed by adding opus to your list of dependencies in mix.exs : def deps do [ { :opus , &quot;~&gt; 0.8&quot; } ] end","ref":"readme.html#installation","title":"Opus - Installation","type":"extras"},{"doc":"hexdocs wiki tutorial","ref":"readme.html#documentation","title":"Opus - Documentation","type":"extras"},{"doc":"Each Opus pipeline module has a single entry point and returns tagged tuples {:ok, value} | {:error, error} A pipeline is a composition of stateless stages A stage returning {:error, _} halts the pipeline A stage may be skipped based on a condition function ( :if and :unless options) Exceptions are converted to {:error, error} tuples by default An exception may be left to raise using the :raise option Each stage of the pipeline is instrumented. Metrics are captured automatically (but can be disabled). Errors are meaningful and predictable","ref":"readme.html#conventions","title":"Opus - Conventions","type":"extras"},{"doc":"defmodule ArithmeticPipeline do use Opus.Pipeline step :add_one , with : &amp; ( &amp;1 + 1 ) check :even? , with : &amp; ( rem ( &amp;1 , 2 ) == 0 ) , error_message : :expected_an_even tee :publish_number , if : &amp; Publisher . publishable? / 1 , raise : [ ExternalError ] step :double , if : :lucky_number? step :divide , unless : :lucky_number? step :randomize , with : &amp; ( &amp;1 * :rand . uniform ) link JSONPipeline def double ( n ) , do : n * 2 def divide ( n ) , do : n / 2 def lucky_number? ( n ) when n in 42 .. 1337 , do : true def lucky_number? ( _ ) , do : false end ArithmeticPipeline . call ( 41 ) # Read this blogpost to get started.","ref":"readme.html#usage","title":"Opus - Usage","type":"extras"},{"doc":"The core aspect of this library is defining pipeline modules. As in the example above you need to add use Opus.Pipeline to turn a module into a pipeline. A pipeline module is a composition of stages executed in sequence.","ref":"readme.html#pipeline","title":"Opus - Pipeline","type":"extras"},{"doc":"There are a few different types of stages for different use-cases. All stage functions, expect a single argument which is provided either from initial call/1 of the pipeline module or the return value of the previous stage. An error value is either :error or {:error, any} and anything else is considered a success value. Step This stage processes the input value and with a success value the next stage is called with that value. With an error value the pipeline is halted and an {:error, any} is returned. Check This stage is intended for validations. This stage calls the stage function and unless it returns true it halts the pipeline. Example: defmodule CreateUserPipeline do use Opus.Pipeline check :valid_params? , with : &amp; match? ( %{ email : email } when is_bitstring ( email ) , &amp;1 ) # other stages to actually create the user end Tee This stage is intended for side effects, such as a notification or a call to an external system where the return value is not meaningful. It never halts the pipeline. Link This stage is to link with another Opus.Pipeline module. It calls call/1 for the provided module. If the module is not an Opus.Pipeline it is ignored. Skip The skip macro can be used for linked pipelines. A linked pipeline may act as a true bypass, based on a condition, expressed as either :if or :unless . When skipped, none of the stages are executed and it returns the input, to be used by any next stages of the caller pipeline. A very common use-case is illustrated in the following example: defmodule RetrieveCustomerInformation do use Opus.Pipeline check :valid_query? link FetchFromCache , if : :cacheable? link FetchFromDatabase , if : :db_backed? step :serialize end With skip it can be written as: defmodule RetrieveCustomerInformation do use Opus.Pipeline check :valid_query? link FetchFromCache link FetchFromDatabase step :serialize end A linked pipeline becomes: defmodule FetchFromCache do use Opus.Pipeline skip :assert_suitable , if : :cacheable? step :retrieve_from_cache end Available options The behaviour of each stage can be configured with any of the available options: :with : The function to call to fulfill this stage. It can be an Atom referring to a public function of the module, an anonymous function or a function reference. :if : Makes a stage conditional, it can be either an Atom referring to a public function of the module, an anonymous function or a function reference. For the stage to be executed, the condition must return true . When the stage is skipped, the input is forwarded to the next step if there's one. :unless : The opposite of the :if option, executes the step only when the callback function returns false . :raise : A list of exceptions to not rescue. Defaults to false which converts all exceptions to {:error, %Opus.PipelineError{}} values halting the pipeline. :error_message : An error message to replace the original error when a stage fails. It can be a String or Atom, which will be used directly in place of the original message, or an anonymous function, which receives the input of the failed stage and must return the error message to be used. :retry_times : How many times to retry a failing stage, before halting the pipeline. :retry_backoff : A backoff function to provide delay values for retries. It can be an Atom referring to a public function in the module, an anonymous function or a function reference. It must return an Enumerable.t yielding at least as many numbers as the retry_times . :instrument? : A boolean which defaults to true . Set to false to skip instrumentation for a stage. Retries defmodule ExternalApiPipeline do use Opus.Pipeline step :http_request , retry_times : 8 , retry_backoff : fn -&gt; linear_backoff ( 10 , 30 ) |&gt; cap ( 100 ) end def http_request ( _input ) do # code for the actual request end end The above module, will retry be retried up to 8 times, each time applying a delay from the next value of the retry_backoff function, which returns a Stream. All the functions from the :retry package will be available to be used in retry_backoff .","ref":"readme.html#stages","title":"Opus - Stages","type":"extras"},{"doc":"You can select the stages of a pipeline to run using call/2 with the :except and :only options. Example: # Runs only the stage with the :validate_params name CreateUserPipeline . call ( params , only : [ :validate_params ] # Runs all the stages except the selected ones CreateUserPipeline . call ( params , except : :send_notification )","ref":"readme.html#stage-filtering","title":"Opus - Stage Filtering","type":"extras"},{"doc":"Instrumentation hooks which can be defined: :pipeline_started : Called before a pipeline module is called :before_stage : Called before each stage :stage_skipped : Called when a conditional stage was skipped :stage_completed : Called after each stage :pipeline_completed : Called after pipeline module has returned You can disable all instrumentation callbacks for a stage using instrument?: false . defmodule ArithmeticPipeline do use Opus.Pipeline step :double , instrument? : false end You can define module specific instrumentation callbacks using: defmodule ArithmeticPipeline do use Opus.Pipeline step :double , with : &amp; ( &amp;1 * 2 ) step :triple , with : &amp; ( &amp;1 * 3 ) instrument :before_stage , fn %{ input : input } -&gt; IO . inspect input end # Will be called only for the matching stage instrument :stage_completed , %{ stage : %{ name : :triple } } , fn %{ time : time } -&gt; # send to the monitoring tool of your choice end end You can define a default instrumentation module for all your pipelines by adding in your config/*.exs : config :opus , :instrumentation , YourModule # but you may choose to provide a list of modules config :opus , :instrumentation , [ YourModuleA , YourModuleB ] An instrumentation module has to export instrument/3 functions like: defmodule CustomInstrumentation do def instrument ( :pipeline_started , %{ pipeline : ArithmeticPipeline } , %{ input : input } ) do # publish the metrics to specific backend end def instrument ( :before_stage , %{ stage : %{ pipeline : pipeline } } , %{ input : input } ) do # publish the metrics to specific backend end def instrument ( :stage_completed , %{ stage : %{ pipeline : ArithmeticPipeline } } , %{ time : time } ) do # publish the metrics to specific backend end def instrument ( :pipeline_completed , %{ pipeline : ArithmeticPipeline } , %{ result : result , time : total_time } ) do # publish the metrics to specific backend end def instrument ( _ , _ , _ ) , do : nil end Telemetry Opus includes an instrumentation module which emits events using the :telemetry library. To enable it, change your config/config.exs with: config :opus , :instrumentation , [ Opus.Telemetry ] Browse the available events here . For instructions to integrate Opus Telemetry metrics in your Phoenix application, read this post .","ref":"readme.html#instrumentation","title":"Opus - Instrumentation","type":"extras"},{"doc":"You may choose to provide some common options to all the stages of a pipeline. :raise : A list of exceptions to not rescue. When set to true , Opus does not handle any exceptions. Defaults to false which converts all exceptions to {:error, %Opus.PipelineError{}} values halting the pipeline. :instrument? : A boolean which defaults to true . Set to false to skip instrumentation for a module. defmodule ArithmeticPipeline do use Opus.Pipeline , instrument? : false , raise : true # The pipeline opts will disable instrumentation for this module # and will not rescue exceptions from any of the stages step :double , with : &amp; ( &amp;1 * 2 ) step :triple , with : &amp; ( &amp;1 * 3 ) end","ref":"readme.html#module-global-options","title":"Opus - Module-Global Options","type":"extras"},{"doc":"You may visualise your pipelines using Opus.Graph : Opus.Graph . generate ( :your_app ) # =&gt; :exclamation: This feature requires the opus_graph package to be installed, add it in your mix.exs. defp deps do { :opus_graph , &quot;~&gt; 0.1&quot; , only : [ :dev ] } end Setup First make sure to add graphvix to your dependencies: # in mix.exs defp deps do [ { :opus , &quot;~&gt; 0.5&quot; } , { :graphvix , &quot;~&gt; 0.5&quot; , only : [ :dev ] } ] end This feature uses graphviz , so make sure to have it installed. To install it: # MacOS brew install graphviz # Debian / Ubuntu apt-get install graphviz Opus.Graph is in fact a pipeline and its visualisation is: You can customise the visualisation: Opus.Graph . generate ( :your_app , %{ filetype : :svg } ) # =&gt; Read the available visualisation options here .","ref":"readme.html#graph","title":"Opus - Graph","type":"extras"},{"doc":"dry.rb - transaction trailblazer - operation","ref":"readme.html#influences","title":"Opus - Influences","type":"extras"},{"doc":"Quiqup Engineering - How to Create Beautify Pipelines with Opus Pagerduty - How I Centralized our Scattered Business Logic Into One Clear Pipeline for our Elixir Webhook Service A Slack bookmarking application in Elixir with Opus Opus Telemetry Using Opus in your company / project? Let us know by submitting an issue describing how you use it.","ref":"readme.html#press","title":"Opus - Press","type":"extras"},{"doc":"Copyright (c) 2018 Dimitris Zorbas, MIT License. See LICENSE.txt for further details.","ref":"readme.html#license","title":"Opus - License","type":"extras"},{"doc":"","ref":"tutorial.html","title":"Tutorial","type":"extras"},{"doc":"Mix . install ( [ { :opus , &quot;~&gt; 0.8&quot; } , { :kino , &quot;~&gt; 0.5&quot; } ] ) Below you'll see a simple pipeline with two steps. It takes a number, adds 1 then multiplies by 2. When a module is made a pipeline with use Opus.Pipeline it can be called with a call/1 function. So our module below can be called with: ArithmeticPipeline . call ( number ) defmodule ArithmeticPipeline do use Opus.Pipeline step ( :add_one , with : &amp; ( &amp;1 + 1 ) ) step ( :multiply_by_two ) def multiply_by_two ( n ) , do : n * 2 end input = Kino.Input . number ( &quot;number&quot; , default : 0 ) number = Kino.Input . read ( input ) ArithmeticPipeline . call ( number )","ref":"tutorial.html#your-first-pipeline","title":"Tutorial - Your first pipeline","type":"extras"},{"doc":"So far we've only defined step stages. This stage processes the input value and with a success value the next stage is called with that value. With an error value the pipeline is halted and an {:error, any} is returned. defmodule ArithmeticPipelineWithErrors do use Opus.Pipeline step ( :add_one , with : &amp; ( &amp;1 + 1 ) ) step ( :multiply_by_two ) step ( :add_three , with : &amp; ( &amp;1 + 3 ) ) def multiply_by_two ( n ) when n &lt; 42 , do : n * 2 def multiply_by_two ( _ ) , do : { :error , &quot;I only handle numbers &lt; 42&quot; } end Try this out with a number &gt; 42 and you should see an Opus.PipelineError . As you can see when a step function returns an error tuple like {:error, &quot;I only handle numbers &lt; 42&quot;} the pipeline is halted an the next stages are not executed. input2 = Kino.Input . number ( &quot;input2&quot; , default : 43 ) number = Kino.Input . read ( input2 ) ArithmeticPipelineWithErrors . call ( number )","ref":"tutorial.html#error-handling","title":"Tutorial - Error Handling","type":"extras"},{"doc":"defmodule Validator do use Opus.Pipeline check ( :valid_user , with : &amp; match? ( %{ user : %{ id : id } } when is_integer ( id ) , &amp;1 ) ) check ( :even_user , with : &amp; ( rem ( &amp;1 . user . id , 2 ) == 0 ) , error_message : &quot;User should have an even id&quot; ) end input3 = Kino.Input . number ( &quot;input3&quot; , default : 3 ) number = Kino.Input . read ( input3 ) Validator . call ( %{ user : %{ id : number } } ) The error message to return when a check fails is configurable. It can be an atom, string or a function. defmodule ValidatorWithError do use Opus.Pipeline check ( :valid_user , with : &amp; match? ( %{ user : %{ id : id } } when is_integer ( id ) , &amp;1 ) ) check ( :even_user , with : &amp; ( rem ( &amp;1 . user . id , 2 ) == 0 ) , error_message : fn %{ user : %{ id : id } } -&gt; &quot;Oh the user should have an even id, \#{ id } is not even&quot; end ) end ValidatorWithError . call ( %{ user : %{ id : 5 } } )","ref":"tutorial.html#validating-input","title":"Tutorial - Validating Input","type":"extras"},{"doc":"You can use the tee macro for side-effects. The return value in such stages is ignored. defmodule ArithmeticSideEffectsPipeline do use Opus.Pipeline step ( :add_one , with : &amp; ( &amp;1 + 1 ) ) tee ( :print_number , with : fn n -&gt; IO . puts ( &quot;The given number is.. \#{ n } &quot; ) # The following error will be ignored raise &quot;error&quot; end ) step ( :multiply_by_two ) def multiply_by_two ( n ) , do : n * 2 end Notice how raising an error does not halt the pipeline with tee . input4 = Kino.Input . number ( &quot;input4&quot; , default : 5 ) number = Kino.Input . read ( input4 ) ArithmeticSideEffectsPipeline . call ( number )","ref":"tutorial.html#side-effects-with-the-tee-stage","title":"Tutorial - Side-effects with the tee stage","type":"extras"},{"doc":"Pipelines can call other pipelines and there's the link macro to make that easier. link is essentially a step where the linked pipeline is called with the step function argument and the step returns the return value of the linked pipeline. input5 = Kino.Input . text ( &quot;input4&quot; , default : &quot;5&quot; ) input6 = Kino.Input . text ( &quot;input4&quot; , default : &quot;6&quot; ) defmodule ReadFirstInput do use Opus.Pipeline step ( :read , with : &amp; put_in ( &amp;1 [ :a ] , Kino.Input . read ( &amp;1 [ :input_a ] ) ) ) end defmodule ReadSecondInput do use Opus.Pipeline step ( :read , with : &amp; put_in ( &amp;1 [ :b ] , Kino.Input . read ( &amp;1 [ :input_b ] ) ) ) end defmodule Calculator do use Opus.Pipeline link ( ReadFirstInput ) link ( ReadSecondInput ) step ( :parse ) step ( :add , with : &amp; ( &amp;1 . a + &amp;1 . b ) , if : &amp; match? ( %{ operation : :add } , &amp;1 ) ) step ( :multiply , with : &amp; ( &amp;1 . a * &amp;1 . b ) , if : &amp; match? ( %{ operation : :multiply } , &amp;1 ) ) def parse ( %{ a : a , b : b } = calculation ) do { a , _ } = Integer . parse ( a ) { b , _ } = Integer . parse ( b ) %{ calculation | a : a , b : b } end end Notice how we leverage the if option to calculate based on the given operation. Both if and unless can be used to make a stage optional. Calculator . call ( %{ operation : :add , input_a : input5 , input_b : input6 } ) Calculator . call ( %{ operation : :multiply , input_a : input5 , input_b : input6 } )","ref":"tutorial.html#linking-pipelines","title":"Tutorial - Linking Pipelines","type":"extras"}]